{#
    epub/layout.html
    ~~~~~~~~~~~~~~~~

    Sphinx layout template for the epub theme.

    :copyright: Copyright 2007-2023 by the Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.
#}
{%- extends "basic/layout.html" %}

{%- block doctype -%}
{{ doctype }}
{%- endblock -%}
{# add only basic navigation links #}
{% block sidebar1 %}{% endblock %}
{% block sidebar2 %}{% endblock %}
{% block relbar1 %}{% endblock %}
{% block relbar2 %}{% endblock %}
{% block linktags %}{% endblock %}
{% block footer %}{% endblock %}
