{#
    bizstyle/layout.html
    ~~~~~~~~~~~~~~~~~~~~

    Sphinx layout template for the bizstyle theme.

    :copyright: Copyright 2007-2023 by Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.
#}
{% extends "basic/layout.html" %}

{%- block scripts %}
    {{ super() }}
    <script src="{{ pathto('_static/bizstyle.js', 1) }}"></script>
{%- endblock %}

{# doctype override #}
{%- block doctype %}
<!doctype html>
{%- endblock %}

{%- block extrahead %}
    <meta name="viewport" content="width=device-width,initial-scale=1.0" />
    <!--[if lt IE 9]>
    <script src="_static/css3-mediaqueries.js"></script>
    <![endif]-->
{%- endblock %}
