{% extends "layout.html" %}
{% set title = _('Sphinx documentation contents') %}
{% block body %}
<h1>{{ _('Sphinx documentation contents') }}</h1>
<div class="toctree-wrapper compound">
{{ toctree(includehidden=True, collapse=False, maxdepth=3) }}
</div>
{% endblock %}
