{% extends "!layout.html" %}
{%- block css %}
  {{ super() }}
    <link rel="stylesheet" href="_static/more_persistent.css" type="text/css" />
    <link rel="stylesheet" href="_static/more_default.css" type="text/css" title="Default" />
    <link rel="alternate stylesheet" href="_static/more_alternate1.css" type="text/css" title="Alternate" />
    <link rel="alternate stylesheet" href="_static/more_alternate2.css" type="text/css" />
{%- endblock %}
