{% extends "layout.html" %}
{% block body %}
  <h1>{{ _('Welcome') }}</h1>
  <p>{%trans%}Sphinx {{ version }}{%endtrans%}</p>
{% endblock %}

{% block comment %}
utf-8 encoded string: ニシキヘビ
{% endblock %}

