from __future__ import annotations

import typing
from typing import final


@typing.final
class Class:
    """docstring"""

    @final
    def meth1(self):
        """docstring"""

    def meth2(self):
        """docstring"""
