Documentation.addTranslations({
    "locale": "pt_PT",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "",
        "&#169; %(copyright_prefix)s %(copyright)s.": "",
        ", in ": ", em",
        "About these documents": "Sobre estes documentos",
        "Automatically generated list of changes in version %(version)s": "Lista de altera\u00e7\u00f5es gerada automaticamente na vers\u00e3o %(version)s",
        "C API changes": "Altera\u00e7\u00f5es na API C",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "",
        "Collapse sidebar": "Recolher painel lateral",
        "Complete Table of Contents": "Tabela de Conte\u00fados Completa",
        "Contents": "Conte\u00fado",
        "Copyright": "Copyright",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "",
        "Expand sidebar": "Expandir painel lateral",
        "Full index on one page": "\u00cdndice completo numa p\u00e1gina",
        "General Index": "\u00cdndice Geral",
        "Global Module Index": "\u00cdndice Global de M\u00f3dulos",
        "Go": "Ir",
        "Hide Search Matches": "Esconder Resultados da Pesquisa",
        "Index": "\u00cdndice",
        "Index &ndash; %(key)s": "\u00cdndice &ndash; %(key)s",
        "Index pages by letter": "Paginas de \u00edndice por letra",
        "Indices and tables:": "\u00cdndices e tabelas:",
        "Last updated on %(last_updated)s.": "\u00daltima actualiza\u00e7\u00e3o em %(last_updated)s.",
        "Library changes": "Altera\u00e7\u00f5es na biblioteca",
        "Navigation": "Navega\u00e7\u00e3o",
        "Next topic": "Pr\u00f3ximo t\u00f3pico",
        "Other changes": "Outras altera\u00e7\u00f5es",
        "Overview": "Vis\u00e3o geral",
        "Please activate JavaScript to enable the search\n    functionality.": "Por favor ligue o JavaScript para habilitar a\nfuncionalidade de pesquisa.",
        "Preparing search...": "A preparar a pesquisa...",
        "Previous topic": "T\u00f3pico anterior",
        "Quick search": "Pesquisa r\u00e1pida",
        "Search": "Pesquisar",
        "Search Page": "P\u00e1gina de Pesquisa",
        "Search Results": "Resultados da Pesquisa",
        "Search finished, found ${resultCount} page(s) matching the search query.": "",
        "Search within %(docstitle)s": "Pesquisar dentro de %(docstitle)s",
        "Searching": "A Pesquisar",
        "Searching for multiple words only shows matches that contain\n    all words.": "",
        "Show Source": "Exibir Fonte",
        "Table of Contents": "",
        "This Page": "Esta P\u00e1gina",
        "Welcome! This is": "Bem Vindo(a)! Esta \u00e9",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "A sua pesquisa n\u00e3o encontrou nenhum documento. Por favor, confirme que todas as palavras est\u00e3o escritas corretamente e que selecionou categorias suficientes.",
        "all functions, classes, terms": "todas as fun\u00e7\u00f5es, classes, termos",
        "can be huge": "pode ser enorme",
        "last updated": "\u00faltima actualiza\u00e7\u00e3o",
        "lists all sections and subsections": "Listar todas as sec\u00e7\u00f5es e subsec\u00e7\u00f5es",
        "next chapter": "pr\u00f3ximo cap\u00edtulo",
        "previous chapter": "cap\u00edtulo anterior",
        "quick access to all modules": "acesso r\u00e1pido a todos os m\u00f3dulos",
        "search": "pesquisar",
        "search this documentation": "Pesquisar esta documenta\u00e7\u00e3o",
        "the documentation for": "a documenta\u00e7\u00e3o de"
    },
    "plural_expr": "(n == 0 || n == 1) ? 0 : n != 0 && n % 1000000 == 0 ? 1 : 2"
});