Documentation.addTranslations({
    "locale": "zh_Hans_CN",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "%(filename)s &#8212; %(docstitle)s",
        "&#169; %(copyright_prefix)s %(copyright)s.": "&#169; %(copyright_prefix)s %(copyright)s.",
        ", in ": "\uff0c\u5728 ",
        "About these documents": "\u5173\u4e8e\u6b64\u6587\u6863",
        "Automatically generated list of changes in version %(version)s": "\u81ea\u52a8\u751f\u6210\u7684 %(version)s \u7248\u672c\u53d8\u66f4\u5217\u8868",
        "C API changes": "C API \u7684\u53d8\u66f4",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "\u4e8e\u7248\u672c %(version)s&#8212; %(docstitle)s \u53d8\u66f4",
        "Collapse sidebar": "\u6298\u53e0\u8fb9\u680f",
        "Complete Table of Contents": "\u5b8c\u6574\u76ee\u5f55",
        "Contents": "\u76ee\u5f55",
        "Copyright": "\u7248\u6743\u6240\u6709",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "\u7531 <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s\u521b\u5efa\u3002",
        "Expand sidebar": "\u5c55\u5f00\u8fb9\u680f",
        "Full index on one page": "\u5355\u9875\u5168\u7d22\u5f15",
        "General Index": "\u603b\u7d22\u5f15",
        "Global Module Index": "\u5168\u5c40\u6a21\u5757\u7d22\u5f15",
        "Go": "\u63d0\u4ea4",
        "Hide Search Matches": "\u9690\u85cf\u641c\u7d22\u7ed3\u679c",
        "Index": "\u7d22\u5f15",
        "Index &ndash; %(key)s": "\u7d22\u5f15 &ndash; %(key)s",
        "Index pages by letter": "\u5b57\u6bcd\u7d22\u5f15",
        "Indices and tables:": "\u7d22\u5f15\u548c\u8868\u683c\uff1a",
        "Last updated on %(last_updated)s.": "\u6700\u540e\u66f4\u65b0\u4e8e %(last_updated)s.",
        "Library changes": "\u5e93\u7684\u53d8\u66f4",
        "Navigation": "\u5bfc\u822a",
        "Next topic": "\u4e0b\u4e00\u4e3b\u9898",
        "Other changes": "\u5176\u4ed6\u53d8\u66f4",
        "Overview": "\u6982\u8ff0",
        "Please activate JavaScript to enable the search\n    functionality.": "\u8bf7\u6fc0\u6d3b JavaScript \u4ee5\u5f00\u542f\u641c\u7d22\u529f\u80fd\u3002",
        "Preparing search...": "\u6b63\u5728\u51c6\u5907\u641c\u7d22\u2026\u2026",
        "Previous topic": "\u4e0a\u4e00\u4e3b\u9898",
        "Quick search": "\u5feb\u901f\u641c\u7d22",
        "Search": "\u641c\u7d22",
        "Search Page": "\u641c\u7d22\u9875\u9762",
        "Search Results": "\u641c\u7d22\u7ed3\u679c",
        "Search finished, found ${resultCount} page(s) matching the search query.": "\u641c\u7d22\u5b8c\u6210\uff0c\u5339\u914d\u5230 ${resultCount} \u9875\u3002",
        "Search within %(docstitle)s": "\u5728 %(docstitle)s \u4e2d\u641c\u7d22",
        "Searching": "\u6b63\u5728\u641c\u7d22\u4e2d",
        "Searching for multiple words only shows matches that contain\n    all words.": "\u5f53\u641c\u7d22\u591a\u4e2a\u5173\u952e\u8bcd\u65f6\uff0c\u53ea\u4f1a\u663e\u793a\u540c\u65f6\u5305\u542b\u6240\u6709\u5173\u952e\u8bcd\u7684\u5185\u5bb9\u3002",
        "Show Source": "\u663e\u793a\u6e90\u4ee3\u7801",
        "Table of Contents": "\u76ee\u5f55",
        "This Page": "\u672c\u9875",
        "Welcome! This is": "\u6b22\u8fce\uff01",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "\u60a8\u7684\u641c\u7d22\u6ca1\u6709\u5339\u914d\u5230\u6587\u6863\u3002\u8bf7\u786e\u4fdd\u5173\u952e\u8bcd\u62fc\u5199\u6b63\u786e\uff0c\u5e76\u4e14\u9009\u62e9\u4e86\u5408\u9002\u7684\u5206\u7c7b\u3002",
        "all functions, classes, terms": "\u6240\u6709\u51fd\u6570\u3001\u7c7b\u3001\u672f\u8bed\u8bcd\u6c47",
        "can be huge": "\u53ef\u80fd\u4f1a\u5927",
        "last updated": "\u6700\u540e\u66f4\u65b0\u4e8e",
        "lists all sections and subsections": "\u5217\u51fa\u6240\u6709\u7684\u7ae0\u8282\u548c\u90e8\u5206",
        "next chapter": "\u4e0b\u4e00\u7ae0",
        "previous chapter": "\u4e0a\u4e00\u7ae0",
        "quick access to all modules": "\u5feb\u901f\u67e5\u770b\u6240\u6709\u7684\u6a21\u5757",
        "search": "\u641c\u7d22",
        "search this documentation": "\u641c\u7d22\u6587\u6863",
        "the documentation for": "\u672c\u6587\u6863\u5c5e\u4e8e"
    },
    "plural_expr": "0"
});