#: docstring
#:
#: :meta hide-value:
SENTINEL1 = object()

#: :meta hide-value:
SENTINEL2 = object()


class Foo:
    """docstring"""

    #: docstring
    #:
    #: :meta hide-value:
    SENTINEL1 = object()

    #: :meta hide-value:
    SENTINEL2 = object()
