class Bar:
    """Bar class"""

    pass


class PublicBar:
    """Public Bar class"""

    pass


def foo():
    """Foo function"""
    pass


def public_foo():
    """Public Foo function"""
    pass
