exclude_patterns = ['_build']

rst_prolog = """\
.. |subst_prolog_1| replace:: prologue substitute text

.. |subst_prolog_2| image:: /img.png
"""

rst_epilog = """\
.. |subst_epilog_1| replace:: epilogue substitute text

.. |subst_epilog_2| image:: /i18n.png
"""
