# Literally included file using Python highlighting

foo = "Including Unicode characters: üöä"  # This will be changed

class FooOrig:
    pass

class BarOrig:
    def baz():
        pass

def bar(): pass
