Documentation.addTranslations({
    "locale": "it",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "%(filename)s &#8212; %(docstitle)s",
        "&#169; %(copyright_prefix)s %(copyright)s.": "",
        ", in ": ", in ",
        "About these documents": "A proposito di questi documenti",
        "Automatically generated list of changes in version %(version)s": "Lista delle modifiche generata automaticamente nella versione %(version)s",
        "C API changes": "Modifiche nelle API C",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "Cambiamenti nella Versione %(version)s &#8212; %(docstitle)s",
        "Collapse sidebar": "Comprimi la barra laterale",
        "Complete Table of Contents": "Tabella dei contenuti completa",
        "Contents": "Contenuti",
        "Copyright": "Copyright",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "",
        "Expand sidebar": "Espandi la barra laterale",
        "Full index on one page": "Indice completo in una pagina",
        "General Index": "Indice generale",
        "Global Module Index": "Indice dei moduli",
        "Go": "Vai",
        "Hide Search Matches": "Nascondi i risultati della ricerca",
        "Index": "Indice",
        "Index &ndash; %(key)s": "Indice &ndash; %(key)s",
        "Index pages by letter": "Indice delle pagine per lettera",
        "Indices and tables:": "Indici e tabelle:",
        "Last updated on %(last_updated)s.": "Ultimo aggiornamento %(last_updated)s.",
        "Library changes": "Modifiche nella libreria",
        "Navigation": "Navigazione",
        "Next topic": "Argomento successivo",
        "Other changes": "Altre modifiche",
        "Overview": "Sintesi",
        "Please activate JavaScript to enable the search\n    functionality.": "Attiva JavaScript per abilitare la funzione\u23ce\ndi ricerca.",
        "Preparing search...": "Preparo la ricerca...",
        "Previous topic": "Argomento precedente",
        "Quick search": "Ricerca veloce",
        "Search": "Cerca",
        "Search Page": "Cerca",
        "Search Results": "Risultati della ricerca",
        "Search finished, found ${resultCount} page(s) matching the search query.": "",
        "Search within %(docstitle)s": "Cerca in %(docstitle)s",
        "Searching": "Cerca",
        "Searching for multiple words only shows matches that contain\n    all words.": "",
        "Show Source": "Mostra sorgente",
        "Table of Contents": "",
        "This Page": "Questa pagina",
        "Welcome! This is": "Benvenuto! Questa \u00e8",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "La tua ricerca non corrisponde a nessun documento. Verifica che tutte le parole siano scritte correttamente e di aver scelto un numero sufficiente di categorie.",
        "all functions, classes, terms": "tutte le funzioni, classi e moduli",
        "can be huge": "pu\u00f2 essere enorme",
        "last updated": "ultimo aggiornamento",
        "lists all sections and subsections": "elenca l'insieme delle sezioni e sottosezioni",
        "next chapter": "capitolo successivo",
        "previous chapter": "capitolo precedente",
        "quick access to all modules": "accesso veloce ai moduli",
        "search": "cerca",
        "search this documentation": "cerca in questa documentazione",
        "the documentation for": "la documentazione per"
    },
    "plural_expr": "n == 1 ? 0 : n != 0 && n % 1000000 == 0 ? 1 : 2"
});