Documentation.addTranslations({
    "locale": "es",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "%(filename)s &#8212; %(docstitle)s",
        "&#169; %(copyright_prefix)s %(copyright)s.": "",
        ", in ": ", en ",
        "About these documents": "Sobre este documento",
        "Automatically generated list of changes in version %(version)s": "Lista de cambios generada autom\u00e1ticamente en la versi\u00f3n %(version)s",
        "C API changes": "Cambios en la API C",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "Cambios en la versi\u00f3n %(version)s &#8212; %(docstitle)s",
        "Collapse sidebar": "Contraer barra lateral",
        "Complete Table of Contents": "\u00cdndice de contenidos completo",
        "Contents": "Contenidos",
        "Copyright": "Copyright",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "Creado usando <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.",
        "Expand sidebar": "Expandir barra lateral",
        "Full index on one page": "\u00cdndice completo en una p\u00e1gina",
        "General Index": "\u00cdndice General",
        "Global Module Index": "\u00cdndice Global de M\u00f3dulos",
        "Go": "Ir a",
        "Hide Search Matches": "Ocultar coincidencias de la b\u00fasqueda",
        "Index": "\u00cdndice",
        "Index &ndash; %(key)s": "\u00cdndice &ndash; %(key)s",
        "Index pages by letter": "\u00cdndice alfab\u00e9tico de p\u00e1ginas",
        "Indices and tables:": "\u00cdndices y tablas:",
        "Last updated on %(last_updated)s.": "Actualizado por \u00faltima vez en %(last_updated)s.",
        "Library changes": "Cambios en la biblioteca",
        "Navigation": "Navegaci\u00f3n",
        "Next topic": "Pr\u00f3ximo tema",
        "Other changes": "Otros cambios",
        "Overview": "Resumen",
        "Please activate JavaScript to enable the search\n    functionality.": "Por favor, active JavaScript para habilitar la funcionalidad\n    de b\u00fasqueda.",
        "Preparing search...": "Preparando b\u00fasqueda...",
        "Previous topic": "Tema anterior",
        "Quick search": "B\u00fasqueda r\u00e1pida",
        "Search": "B\u00fasqueda",
        "Search Page": "P\u00e1gina de B\u00fasqueda",
        "Search Results": "Resultados de la b\u00fasqueda",
        "Search finished, found ${resultCount} page(s) matching the search query.": "B\u00fasqueda finalizada, se encontraron ${resultCount} p\u00e1ginas que coinciden con la consulta de b\u00fasqueda.",
        "Search within %(docstitle)s": "Buscar en %(docstitle)s",
        "Searching": "Buscando",
        "Searching for multiple words only shows matches that contain\n    all words.": "La b\u00fasqueda de varias palabras solo muestra coincidencias que contienen\n    todas las palabras.",
        "Show Source": "Mostrar el c\u00f3digo",
        "Table of Contents": "Tabla de contenido",
        "This Page": "Esta p\u00e1gina",
        "Welcome! This is": "\u00a1Bienvenido! Este es",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "Su b\u00fasqueda no coincide con ning\u00fan documentos. Por favor, aseg\u00farese de que todas las palabras est\u00e9n correctamente escritas y que usted all\u00e1 seleccionado las suficientes categor\u00edas.",
        "all functions, classes, terms": "todas las funciones, clases, t\u00e9rminos",
        "can be huge": "puede ser muy grande",
        "last updated": "actualizado por \u00faltima vez el",
        "lists all sections and subsections": "muestra todas las secciones y subsecciones",
        "next chapter": "pr\u00f3ximo cap\u00edtulo",
        "previous chapter": "cap\u00edtulo anterior",
        "quick access to all modules": "acceso r\u00e1pido a todos los m\u00f3dulos",
        "search": "buscar",
        "search this documentation": "buscar en esta documentaci\u00f3n",
        "the documentation for": "la documentaci\u00f3n para"
    },
    "plural_expr": "n == 1 ? 0 : n != 0 && n % 1000000 == 0 ? 1 : 2"
});