# Literally included file using Python highlighting

foo = "Including Unicode characters: üöä"

class Foo:
    pass

class Bar:
    def baz():
        pass

# comment after Bar class definition
def bar(): pass
