Documentation.addTranslations({
    "locale": "ar",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "",
        "&#169; %(copyright_prefix)s %(copyright)s.": "",
        ", in ": "",
        "About these documents": "",
        "Automatically generated list of changes in version %(version)s": "",
        "C API changes": "",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "",
        "Collapse sidebar": "",
        "Complete Table of Contents": "",
        "Contents": "\u0627\u0644\u0645\u062d\u062a\u0648\u0649",
        "Copyright": "\u0627\u0644\u062d\u0642\u0648\u0642",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "",
        "Expand sidebar": "",
        "Full index on one page": "",
        "General Index": "\u0627\u0644\u0641\u0647\u0631\u0633 \u0627\u0644\u0639\u0627\u0645",
        "Global Module Index": "",
        "Go": "",
        "Hide Search Matches": "",
        "Index": "",
        "Index &ndash; %(key)s": "",
        "Index pages by letter": "",
        "Indices and tables:": "",
        "Last updated on %(last_updated)s.": "",
        "Library changes": "\u062a\u062c\u0647\u064a\u0632 \u0627\u0644\u0628\u062d\u062b",
        "Navigation": "",
        "Next topic": "\u0627\u0644\u0645\u0648\u0636\u0648\u0639 \u0627\u0644\u062a\u0627\u0644\u064a",
        "Other changes": "",
        "Overview": "",
        "Please activate JavaScript to enable the search\n    functionality.": "",
        "Preparing search...": "",
        "Previous topic": "\u0627\u0644\u0645\u0648\u0636\u0648\u0639 \u0627\u0644\u0633\u0627\u0628\u0642",
        "Quick search": "\u0627\u0644\u0628\u062d\u062b \u0627\u0644\u0633\u0631\u064a\u0639",
        "Search": "\u0628\u062d\u062b",
        "Search Page": "\u0635\u0641\u062d\u0629 \u0627\u0644\u0628\u062d\u062b",
        "Search Results": "\u0646\u062a\u0627\u0626\u062c \u0627\u0644\u0628\u062d\u062b",
        "Search finished, found ${resultCount} page(s) matching the search query.": "",
        "Search within %(docstitle)s": "\u0627\u0644\u0628\u062d\u062b \u0636\u0645\u0646 %(docstitle)s",
        "Searching": "",
        "Searching for multiple words only shows matches that contain\n    all words.": "",
        "Show Source": "\u0625\u0638\u0647\u0627\u0631 \u0627\u0644\u0645\u0635\u062f\u0631",
        "Table of Contents": "\u0642\u0627\u0626\u0645\u0629 \u0627\u0644\u0645\u062d\u062a\u0648\u064a\u0627\u062a",
        "This Page": "\u0647\u0630\u0647 \u0627\u0644\u0635\u0641\u062d\u0629",
        "Welcome! This is": "\u0623\u0647\u0644\u0627 \u0648\u0633\u0647\u0644\u0627  \u0647\u0630\u0627",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "",
        "all functions, classes, terms": "",
        "can be huge": "",
        "last updated": "",
        "lists all sections and subsections": "",
        "next chapter": "\u0627\u0644\u0641\u0635\u0644 \u0627\u0644\u062a\u0627\u0644\u064a",
        "previous chapter": "\u0627\u0644\u0642\u0633\u0645 \u0627\u0644\u0633\u0627\u0628\u0642",
        "quick access to all modules": "",
        "search": "\u0628\u062d\u062b",
        "search this documentation": "",
        "the documentation for": "\u0627\u0644\u062a\u0648\u062b\u064a\u0642 \u0644"
    },
    "plural_expr": "n==0 ? 0 : n==1 ? 1 : n==2 ? 2 : n%100>=3 && n%100<=10 ? 3 : n%100>=11 && n%100<=99 ? 4 : 5"
});