Documentation.addTranslations({
    "locale": "is",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "",
        "&#169; %(copyright_prefix)s %(copyright)s.": "",
        ", in ": "",
        "About these documents": "",
        "Automatically generated list of changes in version %(version)s": "",
        "C API changes": "",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "",
        "Collapse sidebar": "",
        "Complete Table of Contents": "Heildarefnisyfirlit",
        "Contents": "",
        "Copyright": "",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "",
        "Expand sidebar": "",
        "Full index on one page": "",
        "General Index": "Almennt yfirlit",
        "Global Module Index": "",
        "Go": "",
        "Hide Search Matches": "Fela leitarni\u00f0urst\u00f6\u00f0ur",
        "Index": "Yfirlit",
        "Index &ndash; %(key)s": "",
        "Index pages by letter": "",
        "Indices and tables:": "Atri\u00f0askr\u00e1 og t\u00f6flur:",
        "Last updated on %(last_updated)s.": "",
        "Library changes": "",
        "Navigation": "",
        "Next topic": "N\u00e6sta efni",
        "Other changes": "",
        "Overview": "",
        "Please activate JavaScript to enable the search\n    functionality.": "",
        "Preparing search...": "Undirb\u00fdr leit...",
        "Previous topic": "Fyrra efni",
        "Quick search": "Fl\u00fdtileit",
        "Search": "Leit",
        "Search Page": "Leitars\u00ed\u00f0a",
        "Search Results": "Leitarni\u00f0urst\u00f6\u00f0ur",
        "Search finished, found ${resultCount} page(s) matching the search query.": "",
        "Search within %(docstitle)s": "Leita \u00ed %(docstitle)s",
        "Searching": "Leitar",
        "Searching for multiple words only shows matches that contain\n    all words.": "",
        "Show Source": "",
        "Table of Contents": "Efnisyfirlit",
        "This Page": "",
        "Welcome! This is": "",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "Leitin bar ekki \u00e1rangur. Vinsamlegast gangi\u00f0 \u00far skugga um a\u00f0 or\u00f0in s\u00e9u r\u00e9tt skrifu\u00f0 og a\u00f0 \u00fe\u00fa hafir vali\u00f0 n\u00e6gilega marga flokka.",
        "all functions, classes, terms": "",
        "can be huge": "",
        "last updated": "",
        "lists all sections and subsections": "",
        "next chapter": "n\u00e6sti kafli",
        "previous chapter": "fyrri kafli",
        "quick access to all modules": "",
        "search": "leita",
        "search this documentation": "leita",
        "the documentation for": ""
    },
    "plural_expr": "(n % 10 != 1 || n % 100 == 11)"
});