latex_documents = [
    ('index', 'test.tex', 'The basic Sphinx documentation for testing', 'Sphinx', 'report')
]
html_logo = "https://www.python.org/static/img/python-logo.png"
html_favicon = "https://www.python.org/static/favicon.ico"
